import os
import asyncio
import re
import sys
import json
import logging
import base64
import yaml
from collections import OrderedDict
from datetime import datetime, timedelta, timezone, date
import random
import string
from typing import Dict, Any, Optional, List, Union
from jinja2 import Environment, StrictUndefined, BaseLoader

# logger = setup_logger(__name__, include_location=True)
#===================================
#  custom logger
#===================================

SUCCESS_LEVEL = 25
LOG_SEVERITY = {
    "DEBUG": "🔍",
    "INFO": "ℹ️",
    "SUCCESS": "✅",
    "WARNING": "⚠️",
    "ERROR": "❌",
    "CRITICAL": "🔥"
}

logging.addLevelName(SUCCESS_LEVEL, "SUCCESS")

class CustomLogger(logging.Logger):
    def success(self, message, *args, **kwargs):
        if self.isEnabledFor(SUCCESS_LEVEL):
            self.log(SUCCESS_LEVEL, message, *args, **kwargs, stacklevel=2)

def stringify_extra(value):
    if isinstance(value, (list, dict)):
        return str(value)
    else:
        return value


class CustomFormatter(logging.Formatter):

    def __init__(self, fmt="%(message)s", include_location=False, highlight_scope=True):
        super().__init__(fmt)
        self.include_location = include_location
        self.highlight_scope = highlight_scope

    def format(self, record):
        icon = LOG_SEVERITY.get(record.levelname, "ℹ️")
        if hasattr(record, "scope"):
            scope_highlight = f"\033[32m{record.scope}\033[0m"
        else:
            scope_highlight = ""
        location = ""
        if self.include_location and hasattr(record, "module") and hasattr(record, "funcName") and hasattr(record,
                                                                                                           "lineno"):
            location = f"\033[1;33m({record.module}:{record.funcName}:{record.lineno})\033[0m"

        metadata_line = f"{icon} [{record.levelname}] {scope_highlight} {location}".strip()

        if isinstance(record.msg, (dict, list)):
            message = str(record.msg)
        else:
            message = str(record.msg)

        message_split = message.splitlines()
        if len(message_split) > 1:
            message_line = f"     Message: {message_split[0]}"
            for line in message_split[1:]:
                message_line += f"\n             {line}"
        else:
            message_line = f"     Message: {message}" #.strip()

        extra_items = [
            f"{key}: {stringify_extra(value)}"
            for key, value in record.__dict__.items()
            if key not in [
                "msg", "args", "levelname", "levelno", "pathname",
                "filename", "module", "exc_info", "exc_text", "stack_info",
                "lineno", "funcName", "created", "msecs", "relativeCreated",
                "thread", "threadName", "process", "processName", "scope", "name", "taskName"
            ]
        ]
        extra_info = ""
        if extra_items:
            extra_info = f"\n     {' '.join(extra_items)}"
        formatted_log = f"{metadata_line}\n{message_line}{extra_info}"
        return formatted_log


class JSONFormatter(logging.Formatter):
    def format(self, record):
        log_dict = {
            "level": record.levelname,
            "message": record.msg,
            "time": self.formatTime(record, self.datefmt),
        }
        if hasattr(record, "scope"):
            log_dict["scope"] = record.scope
        if hasattr(record, "module") and hasattr(record, "funcName") and hasattr(record, "lineno"):
            log_dict["location"] = {
                "module": record.module,
                "function": record.funcName,
                "line": record.lineno,
            }
        return json.dumps(log_dict, ensure_ascii=False)

def setup_logger(name: str, include_location=False, use_json=False):
    logging.setLoggerClass(CustomLogger)
    logger = logging.getLogger(name)

    if not any(isinstance(h, logging.StreamHandler) for h in logger.handlers):
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setLevel(logging.DEBUG)
        if use_json:
            stream_handler.setFormatter(JSONFormatter())
        else:
            stream_handler.setFormatter(CustomFormatter(include_location=include_location))
        logger.addHandler(stream_handler)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False
    return logger



logger = setup_logger(__name__, include_location=True)

#===================================
#  jinja2 template rendering
#===================================

def render_template(env: Environment, template: Any, context: Dict, rules: Dict = None) -> Any:
    """
    Render a template using the Jinja2 environment.

    Args:
        env: The Jinja2 environment
        template: The template to render
        context: The context to use for rendering
        rules: Additional rules for rendering

    Returns:
        The rendered template
    """
    if isinstance(template, str) and '{{' in template and '}}' in template:
        logger.debug(f"Render template: {template}")
        logger.debug(f"Render template context keys: {list(context.keys())}")
        if 'city' in context:
            logger.debug(f"Render template city value: {context['city']}, Type: {type(context['city'])}")
        if rules:
            logger.debug(f"Render template rules: {rules}")

    render_ctx = dict(context)
    if rules:
        render_ctx.update(rules)

    if isinstance(template, str) and '{{' in template and '}}' in template:
        try:
            expr = template.strip()
            if expr == '{{}}':
                return ""
            if expr.startswith('{{') and expr.endswith('}}'):
                var_path = expr[2:-2].strip()
                if not any(op in var_path for op in ['==', '!=', '<', '>', '+', '-', '*', '/', '|', ' if ', ' else ']):
                    if '.' not in var_path and var_path.strip() in render_ctx:
                        return render_ctx.get(var_path.strip())
                    elif '.' in var_path:
                        parts = var_path.split('.')
                        value = render_ctx
                        valid_path = True
                        for part in parts:
                            part = part.strip()
                            if isinstance(value, dict) and part in value:
                                value = value.get(part)
                            else:
                                valid_path = False
                                break
                        if valid_path:
                            return value

            template_obj = env.from_string(template)
            try:
                rendered = template_obj.render(**render_ctx)
            except Exception as e:
                logger.error(f"Template rendering error: {e}, template: {template}")
                return None

            if (rendered.startswith('[') and rendered.endswith(']')) or \
                    (rendered.startswith('{') and rendered.endswith('}')):
                try:
                    return json.loads(rendered)
                except json.JSONDecodeError:
                    pass

            if rendered.strip() == "":
                return ""

            return rendered
        except Exception as e:
            logger.error(f"Template rendering error: {e}, template: {template}")
            return ""
    elif isinstance(template, dict):
        if not template:
            return template
        return {k: render_template(env, v, render_ctx, rules) for k, v in template.items()}
    elif isinstance(template, list):
        return [render_template(env, item, render_ctx, rules) for item in template]
    return template

def quote_jinja2_expressions(yaml_text):
    """
    Jinja2 expressions to double quotes.
    """
    jinja_expr_pattern = re.compile(r'''
        ^(\s*[^:\n]+:\s*)         # YAML key and colon with optional indent
        (?!["'])                  # Not already quoted
        (.*{{.*}}.*?)             # Contains Jinja2 template
        (?<!["'])\s*$             # Not ending with a quote
    ''', re.VERBOSE)

    def replacer(match):
        key_part = match.group(1)
        value_part = match.group(2).strip()
        return f'{key_part}"{value_part}"'

    fixed_lines = []
    for line in yaml_text.splitlines():
        fixed_line = jinja_expr_pattern.sub(replacer, line)
        fixed_lines.append(fixed_line)
    return "\n".join(fixed_lines)


#===================================
#  time calendar (დრო)
#===================================

def generate_id() -> str:
    start_date_time = datetime.now()
    timestamp = start_date_time.strftime("%Y%m%d_%H%M%S")
    microseconds = f"{start_date_time.microsecond:06d}"
    suffix = ''.join(random.choices(string.ascii_uppercase + string.digits, k=4))
    return f"{timestamp}_{microseconds}_{suffix}"

def duration_seconds(start_date, end_date) -> int | None:
    return int((start_date - end_date).total_seconds())

def now_utc() -> datetime:
    return datetime.now(timezone.utc)

def today_utc() -> datetime:
    now = now_utc()
    return now.replace(hour=0, minute=0, second=0, microsecond=0)

def week_start_utc(reference: datetime = None) -> datetime:
    if reference is None:
        reference = now_utc()
    monday = reference - timedelta(days=reference.weekday())
    return monday.replace(hour=0, minute=0, second=0, microsecond=0)

def add_days(base: datetime, days: int) -> datetime:
    return base + timedelta(days=days)

def add_weeks(base: datetime, weeks: int) -> datetime:
    return base + timedelta(weeks=weeks)

def month_end(date: datetime) -> datetime:
    next_month = date.replace(day=28) + timedelta(days=4)
    last_day = next_month - timedelta(days=next_month.day)
    return last_day.replace(hour=23, minute=59, second=59, microsecond=999999)

def format_iso8601(dt: datetime) -> str:
    return dt.astimezone(timezone.utc).isoformat()

def format_utc(dt: datetime) -> str:
    return dt.strftime('%Y-%m-%d %H:%M:%S UTC')

def parse_iso8601(iso_str: str) -> datetime:
    return datetime.fromisoformat(iso_str)

def days_between(start: datetime, end: datetime) -> int:
    delta = end - start
    return delta.days

def is_weekend(date: datetime) -> bool:
    return date.weekday() >= 5

def next_weekday(date: datetime, weekday: int) -> datetime:
    days_ahead = weekday - date.weekday()
    if days_ahead <= 0:
        days_ahead += 7
    return date + timedelta(days=days_ahead)

#===================================
# environment
#===================================


def log_level() -> int:
    log_level: int = getattr(logging, os.environ.get("LOG_LEVEL", "INFO").upper(), logging.INFO)
    return log_level

def log_level_name() -> str:
    return logging.getLevelName(log_level())

def is_on(value: str) -> bool:
    return str(value).lower() in ("true", "1", "yes", "on", "y", "t")

def is_off(value: str) -> bool:
    return str(value).lower() in ("false", "0", "no", "off", "n", "f")

async def mkdir(dir_path):
    try:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, lambda: os.makedirs(dir_path, exist_ok=True))
        return {f"Directory created": {dir_path}}
    except Exception as e:
        raise f"Failed to create directory: {dir_path}. Error: {e}"

#===================================
# serialization
#===================================


def make_serializable(value):
    if isinstance(value, datetime):
        return value.isoformat()
    if isinstance(value, Exception):
        return str(value)
    if isinstance(value, dict):
        return {k: make_serializable(v) for k, v in value.items()}
    if isinstance(value, list):
        return [make_serializable(v) for v in value]
    if hasattr(value, "__dict__"):
        return {k: make_serializable(v) for k, v in value.__dict__.items()}
    return value


class SafeEncoder(json.JSONEncoder):
    def default(self, obj):
        return None

class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, date):
            return obj.isoformat()
        return super().default(obj)

def ordered_yaml_dump(data):
    def convert_ordered_dict(obj):
        if isinstance(obj, OrderedDict):
            return {k: convert_ordered_dict(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [convert_ordered_dict(i) for i in obj]
        return obj

    cleaned_data = convert_ordered_dict(data)
    return yaml.dump(cleaned_data, default_flow_style=False, sort_keys=False)

def ordered_yaml_load(stream):
    class OrderedLoader(yaml.SafeLoader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return OrderedDict(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping
    )
    return yaml.load(stream, OrderedLoader)

def encode_version(version: str) -> str:
    major, minor, patch = map(int, version.split("."))
    return f"{major:03d}.{minor:03d}.{patch:03d}"

def decode_version(encoded_version: str) -> str:
    major, minor, patch = encoded_version.split(".")
    return f"{int(major)}.{int(minor)}.{int(patch)}"

def increment_version(version: str) -> str:
    major, minor, patch = map(int, version.split("."))

    patch += 1
    if patch > 999:
        patch = 0
        minor += 1
        if minor > 999:
            minor = 0
            major += 1
            if major > 999:
                raise ValueError("Version overflow -> version limit is '999.999.999'")

    return f"{major}.{minor}.{patch}"

#===================================
# merging dictionaries and lists
#===================================

def deep_merge(dest: Union[Dict, List, Any], source: Union[Dict, List, Any]) -> Union[Dict, List, Any]:
    if isinstance(dest, dict) and isinstance(source, dict):
        result = dest.copy()
        for key, value in source.items():
            if key in result and isinstance(result[key], (dict, list)) and isinstance(value, (dict, list)):
                result[key] = deep_merge(result[key], value)
            else:
                result[key] = value
        return result
    elif isinstance(dest, list) and isinstance(source, list):
        if all(isinstance(item, dict) for item in dest + source):
            common_keys = set()
            for item in dest + source:
                common_keys.update(item.keys())
            if 'name' in common_keys:
                result = dest.copy()
                source_names = {item.get('name'): item for item in source if 'name' in item}
                for i, item in enumerate(result):
                    if 'name' in item and item['name'] in source_names:
                        result[i] = deep_merge(item, source_names[item['name']])
                        del source_names[item['name']]

                result.extend(source_names.values())
                return result

        result = dest.copy()
        result.extend(source)
        return result
    else:
        return source

#===================================
# connections
#===================================

def get_pgdb_connection(
    db_name: str = None,
    user: str = None,
    password: str = None,
    host: str = None,
    port: str = None,
    schema: str = None
) -> str:
    db_name = db_name or os.environ.get('POSTGRES_DB', 'noetl')
    user = user or os.environ.get('NOETL_USER', 'noetl')
    password = password or os.environ.get('NOETL_PASSWORD', 'noetl')
    host = host or os.environ.get('POSTGRES_HOST', 'localhost')
    port = port or os.environ.get('POSTGRES_PORT', '5434')
    schema = schema or os.environ.get('NOETL_SCHEMA', 'noetl')

    return f"dbname={db_name} user={user} password={password} host={host} port={port} options='-c search_path={schema}'"



import uvicorn
import typer
import os
import json
import logging
import base64
import requests
from pathlib import Path
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from fastapi.responses import HTMLResponse, FileResponse
from noetl.server import router as server_router
from noetl.common import deep_merge, setup_logger
from noetl.worker import NoETLAgent
from noetl.schema import DatabaseSchema

# Setup logger for the main module
logger = setup_logger(__name__, include_location=True)

# Use a distinct name for the Typer CLI application to avoid confusion with the FastAPI app.
cli_app = typer.Typer()


def create_app() -> FastAPI:
    """
    Creates and configures the main FastAPI application instance.
    This function is the factory for Uvicorn.
    """
    app = FastAPI(
        title="NoETL API",
        description="NoETL API server",
        version="0.1.21"
    )

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # --- Database Initialization ---
    try:
        logger.info("Initializing NoETL system metadata.")
        db_schema = DatabaseSchema(auto_setup=False)
        db_schema.create_noetl_metadata()
        db_schema.init_database()
        logger.info("NoETL user and database schema initialized.")
    except Exception as e:
        logger.error(f"Error initializing NoETL system metadata: {e}", exc_info=True)
        logger.warning("Continuing with server startup despite database initialization error.")

    # --- UI Serving Setup ---
    package_dir = Path(__file__).parent
    ui_path = package_dir / "ui"
    templates_path = ui_path / "templates"
    static_path = ui_path / "static"
    assets_path = static_path / "assets"

    templates = None
    if templates_path.exists() and assets_path.exists():
        templates = Jinja2Templates(directory=templates_path)

        # Mount the /assets directory to serve JS, CSS, etc.
        app.mount("/assets", StaticFiles(directory=assets_path), name="assets")
        logger.info(f"UI assets mounted from: {assets_path}")

        # Add a specific route for the favicon
        @app.get("/favicon.svg", include_in_schema=False)
        async def favicon():
            favicon_file = static_path / "favicon.svg"
            if favicon_file.exists():
                return FileResponse(favicon_file)
            return HTMLResponse(status_code=404)
    else:
        logger.warning(f"UI files not found. Searched in: {ui_path}")

    # --- API Router ---
    app.include_router(server_router)

    # --- Health Check ---
    @app.get("/health", include_in_schema=False)
    async def health():
        return {"status": "ok"}

    # --- SPA Catch-all Route ---
    # This single route handles serving all your HTML pages.
    # It must be defined *after* all other API routes.
    if templates:
        @app.get("/{full_path:path}", response_class=HTMLResponse, include_in_schema=False)
        async def serve_spa(request: Request, full_path: str):
            """Serves the appropriate HTML template for any given path."""
            if full_path.startswith("editor") or full_path.startswith("playbook"):
                template_name = "editor.html"
            elif full_path.startswith("execution"):
                template_name = "execution.html"
            elif full_path.startswith("catalog"):
                template_name = "catalog.html"
            else:
                template_name = "index.html"  # Default/fallback

            return templates.TemplateResponse(template_name, {"request": request})
    else:
        @app.get("/", response_class=HTMLResponse, include_in_schema=False)
        async def root_no_ui():
            return {"message": "NoETL API is running, but UI is not available"}

    return app


# --- Typer CLI Commands ---

@cli_app.command("server")
def run_server(
    host: str = typer.Option("0.0.0.0", help="Server host."),
    port: int = typer.Option(8080, help="Server port."),
    reload: bool = typer.Option(False, help="Server auto-reload.")
):
    """Starts the NoETL web server."""
    logger.info(f"Starting NoETL API server at http://{host}:{port}")
    # Use the factory pattern for Uvicorn, which is best practice.
    uvicorn.run("noetl.main:create_app", factory=True, host=host, port=port, reload=reload)


@cli_app.command("agent")
def run_agent(
    file: str = typer.Option(..., "--file", "-f", help="Path to playbook YAML file."),
    mock: bool = typer.Option(False, help="Run in mock mode"),
    output: str = typer.Option("json", "--output", "-o", help="Output format, json or plain."),
    export: str = typer.Option(None, help="Export execution data to Parquet file."),
    mlflow: bool = typer.Option(False, help="Use ML model for workflow control."),
    postgres: str = typer.Option(None, help="Postgres connection string."),
    duckdb: str = typer.Option(None, help="Path to DuckDB file for business logic in playbooks."),
    input: str = typer.Option(None, help="Path to the input payload JSON file for the playbook."),
    payload: str = typer.Option(None, help="JSON input payload string for the playbook."),
    merge: bool = typer.Option(False, help="Merge the input payload with the workload section."),
    debug: bool = typer.Option(False, help="Debug logging mode.")
):
    """Executes a NoETL playbook as a local agent."""
    logging.basicConfig(
        format='[%(levelname)s] %(asctime)s,%(msecs)03d (%(name)s:%(funcName)s:%(lineno)d) - %(message)s',
        datefmt='%Y-%m-%dT%H:%M:%S',
        level=logging.DEBUG if debug else logging.INFO
    )

    try:
        input_payload = None
        if input:
            try:
                with open(input, 'r') as f:
                    input_payload = json.load(f)
                logger.info(f"Loaded input payload from {input}")
            except Exception as e:
                logger.error(f"Error loading input payload: {e}")
                raise typer.Exit(code=1)
        elif payload:
            try:
                input_payload = json.loads(payload)
                logger.info("Parsed input payload from command line")
            except Exception as e:
                logger.error(f"Error parsing payload JSON: {e}")
                raise typer.Exit(code=1)
        pgdb_conn = postgres or os.environ.get("NOETL_PGDB")
        if not pgdb_conn:
            pgdb_conn = f"dbname={os.environ.get('POSTGRES_DB', 'demo_noetl')} user={os.environ.get('POSTGRES_USER', 'demo')} password={os.environ.get('POSTGRES_PASSWORD', 'demo')} host={os.environ.get('POSTGRES_HOST', 'localhost')} port={os.environ.get('POSTGRES_PORT', '5434')}"
            logger.info(f"Using default Postgres connection string: {pgdb_conn}")

        if duckdb:
            os.environ['DUCKDB_PATH'] = duckdb
            logger.info(f"Using DuckDB for business logic: {duckdb}")

        agent = NoETLAgent(file, mock_mode=mock, pgdb=pgdb_conn)
        workload = agent.playbook.get('workload', {})

        if input_payload:
            if merge:
                logger.info("Merge mode: deep merging input payload with workload.")
                merged_workload = deep_merge(workload, input_payload)
                for key, value in merged_workload.items():
                    agent.update_context(key, value)
                agent.update_context('workload', merged_workload)
                agent.store_workload(merged_workload)
            else:
                logger.info("Override mode: replacing the matching workload keys with input payload.")
                new_workload = workload.copy()
                for key, value in input_payload.items():
                    new_workload[key] = value
                for key, value in new_workload.items():
                    agent.update_context(key, value)
                agent.update_context('workload', new_workload)
                agent.store_workload(new_workload)
        else:
            logger.info("Using default workload from playbook.")
            for key, value in workload.items():
                agent.update_context(key, value)
            agent.update_context('workload', workload)
            agent.store_workload(workload)

        results = agent.run(mlflow=mlflow)

        if export:
            agent.export_execution_data(export)

        if output == "json":
            logger.info(json.dumps(results, indent=2))
        else:
            for step, result in results.items():
                logger.info(f"{step}: {result}")

        logger.info(f"Postgres connection: {agent.pgdb}")
        logger.info(f"Open notebook/agent_mission_report.ipynb and set 'pgdb' to {agent.pgdb}")

    except Exception as e:
        logger.error(f"Error executing playbook: {e}", exc_info=True)
        print(f"Error executing playbook: {e}")
        raise typer.Exit(code=1)


@cli_app.command("playbook")
def manage_playbook(
    register: str = typer.Option(None, "--register", "-r", help="Path to playbook file to register."),
    execute: bool = typer.Option(False, "--execute", "-e", help="Execute a playbook by path."),
    path: str = typer.Option(None, "--path", help="Path of the playbook to execute."),
    version: str = typer.Option(None, "--version", "-v", help="Version of the playbook to execute."),
    input: str = typer.Option(None, "--input", "-i", help="Path to payload file."),
    payload: str = typer.Option(None, "--payload", help="Payload string."),
    host: str = typer.Option("localhost", "--host", help="NoETL server host for client connections."),
    port: int = typer.Option(8082, "--port", "-p", help="NoETL server port."),
    sync: bool = typer.Option(True, "--sync", help="Sync up execution data to Postgres."),
    merge: bool = typer.Option(False, "--merge", help="Merge the input payload with the workload section of playbook.")
):
    """Registers or executes a playbook via the NoETL server."""
    if register:
        try:
            if not os.path.exists(register):
                logger.error(f"File not found: {register}")
                raise typer.Exit(code=1)
            with open(register, 'r') as f:
                content = f.read()
            content_base64 = base64.b64encode(content.encode('utf-8')).decode('utf-8')
            url = f"http://{host}:{port}/catalog/register"
            headers = {"Content-Type": "application/json"}
            data = {"content_base64": content_base64}
            logger.info(f"Registering playbook {register} with NoETL server at {url}")
            response = requests.post(url, headers=headers, json=data)

            if response.status_code == 200:
                result = response.json()
                logger.info(f"Playbook registered successfully: {result}")
                logger.info(f"Resource path: {result.get('resource_path')}")
                logger.info(f"Resource version: {result.get('resource_version')}")
            else:
                logger.error(f"Failed to register playbook: {response.status_code}")
                logger.error(f"Response: {response.text}")
                raise typer.Exit(code=1)

        except Exception as e:
            logger.error(f"Error registering playbook: {e}")
            raise typer.Exit(code=1)
    elif execute:
        try:
            if not path:
                logger.error("Path is required when using --execute")
                logger.info("Example: noetl playbook --execute --path weather_example --version 0.1.0")
                raise typer.Exit(code=1)

            input_payload = {}
            if input:
                try:
                    with open(input, 'r') as f:
                        input_payload = json.load(f)
                    logger.info(f"Loaded input payload from {input}")
                except Exception as e:
                    logger.error(f"Error loading input payload from file: {e}")
                    raise typer.Exit(code=1)
            elif payload:
                try:
                    input_payload = json.loads(payload)
                    logger.info("Parsed input payload from command line")
                except Exception as e:
                    logger.error(f"Error parsing payload JSON: {e}")
                    raise typer.Exit(code=1)

            url = f"http://{host}:{port}/agent/execute"
            headers = {"Content-Type": "application/json"}
            data = {
                "path": path,
                "input_payload": input_payload,
                "sync_to_postgres": sync,
                "merge": merge
            }

            if version:
                data["version"] = version

            if version:
                logger.info(f"Executing playbook {path} version {version} on NoETL server at {url}")
            else:
                logger.info(f"Executing playbook {path} (latest version) on NoETL server at {url}")
            response = requests.post(url, headers=headers, json=data)

            if response.status_code == 200:
                result = response.json()
                logger.info(f"Playbook executed.")

                if result.get("status") == "success":
                    logger.info(f"Execution ID: {result.get('execution_id')}")
                    logger.info(f"Result: {json.dumps(result.get('result'), indent=2)}")
                else:
                    logger.error(f"Execution failed: {result.get('error')}.")
                    raise typer.Exit(code=1)
            else:
                logger.error(f"Failed to execute playbook: {response.status_code}.")
                logger.error(f"Response: {response.text}.")
                raise typer.Exit(code=1)

        except Exception as e:
            logger.error(f"Error executing playbook: {e}.")
            raise typer.Exit(code=1)
    else:
        logger.info("No action specified. Use --register to register a playbook or --execute to execute a playbook.")
        logger.info("Examples:")
        logger.info("  noetl playbook --register ./playbook/weather_loop_example.yaml")
        logger.info("  noetl playbook --execute --path weather_example --version 0.1.0 --payload '{\"city\": \"New York\"}'")
        logger.info("  noetl playbook --execute --path weather_example --input ./data/input/payload.json")


def main():
    """Main entry point for the CLI application."""
    cli_app()


# Create the app instance for external imports (e.g., from entry points)
app = create_app()


if __name__ == "__main__":
    main()



import os
import json
import yaml
import tempfile
import psycopg
from typing import Dict, Any, Optional, List
from fastapi import APIRouter, Depends, HTTPException, Request, BackgroundTasks
from fastapi.responses import JSONResponse
from noetl.common import setup_logger, deep_merge, get_pgdb_connection
from noetl.worker import NoETLAgent
from noetl.broker import Broker

logger = setup_logger(__name__, include_location=True)

router = APIRouter()

class CatalogService:
    def __init__(self, pgdb_conn_string: str | None = None):
        self.pgdb_conn_string = pgdb_conn_string if pgdb_conn_string else get_pgdb_connection()

    def get_latest_version(self, resource_path: str) -> str:
        conn = None
        try:
            conn = psycopg.connect(self.pgdb_conn_string)

            with conn.cursor() as cursor:
                cursor.execute(
                    "SELECT COUNT(*) FROM catalog WHERE resource_path = %s",
                    (resource_path,)
                )
                count = cursor.fetchone()[0]
                logger.debug(f"Found {count} entries for resource_path '{resource_path}'")

                if count == 0:
                    logger.debug(f"No entries found for resource_path '{resource_path}', returning default version '0.1.0'")
                    return "0.1.0"

                cursor.execute(
                    "SELECT resource_version FROM catalog WHERE resource_path = %s",
                    (resource_path,)
                )
                versions = [row[0] for row in cursor.fetchall()]
                logger.debug(f"All versions for resource_path '{resource_path}': {versions}")

                cursor.execute(
                    """
                    WITH parsed_versions AS (
                        SELECT 
                            resource_version,
                            CAST(SPLIT_PART(resource_version, '.', 1) AS INTEGER) AS major,
                            CAST(SPLIT_PART(resource_version, '.', 2) AS INTEGER) AS minor,
                            CAST(SPLIT_PART(resource_version, '.', 3) AS INTEGER) AS patch
                        FROM catalog
                        WHERE resource_path = %s
                    )
                    SELECT resource_version
                    FROM parsed_versions
                    ORDER BY major DESC, minor DESC, patch DESC
                    LIMIT 1
                    """,
                    (resource_path,)
                )
                result = cursor.fetchone()

                if result:
                    latest_version = result[0]
                    logger.debug(f"Latest version for resource_path '{resource_path}': '{latest_version}'")
                    return latest_version

                logger.debug(f"No valid version found for resource_path '{resource_path}', returning default version '0.1.0'")
                return "0.1.0"
        except Exception as e:
            logger.exception(f"Error getting latest version for resource_path '{resource_path}': {e}")
            return "0.1.0"
        finally:
            if conn:
                try:
                    conn.close()
                except Exception as close_error:
                    logger.exception(f"Error closing connection: {close_error}")

    def fetch_entry(self, path: str, version: str) -> Optional[Dict[str, Any]]:
        conn = None
        try:
            conn = psycopg.connect(self.pgdb_conn_string)

            with conn.cursor() as cursor:
                cursor.execute(
                    """
                    SELECT resource_path, resource_type, resource_version, content, payload, meta
                    FROM catalog
                    WHERE resource_path = %s AND resource_version = %s
                    """,
                    (path, version)
                )

                result = cursor.fetchone()

                if not result and '/' in path:
                    filename = path.split('/')[-1]
                    logger.info(f"Path not found. Trying to match filename: {filename}")

                    cursor.execute(
                        """
                        SELECT resource_path, resource_type, resource_version, content, payload, meta
                        FROM catalog
                        WHERE resource_path = %s AND resource_version = %s
                        """,
                        (filename, version)
                    )

                    result = cursor.fetchone()

            if result:
                return {
                    "resource_path": result[0],
                    "resource_type": result[1],
                    "resource_version": result[2],
                    "content": result[3],
                    "payload": result[4],
                    "meta": result[5]
                }
            return None

        except Exception as e:
            logger.exception(f"Error fetching catalog entry: {e}.")
            return None
        finally:
            if conn:
                try:
                    conn.close()
                except Exception as close_error:
                    logger.exception(f"Error closing connection: {close_error}")

    def increment_version(self, version: str) -> str:
        try:
            parts = version.split('.')
            while len(parts) < 3:
                parts.append('0')
            major, minor, patch = map(int, parts[:3])
            patch += 1
            return f"{major}.{minor}.{patch}"
        except Exception as e:
            logger.exception(f"Error incrementing version: {e}")
            return f"{version}.1"

    def register_resource(self, content: str, resource_type: str = "playbook") -> Dict[str, Any]:
        conn = None
        try:
            conn = psycopg.connect(self.pgdb_conn_string)
            resource_data = yaml.safe_load(content)
            resource_path = resource_data.get("path", resource_data.get("name", "unknown"))
            logger.debug(f"Extracted resource_path: '{resource_path}'")
            latest_version = self.get_latest_version(resource_path)
            logger.debug(f"Latest version for resource_path '{resource_path}': '{latest_version}'")
            resource_version = self.increment_version(latest_version)
            logger.debug(f"Incremented version: '{resource_version}'")
            with conn.cursor() as cursor:
                cursor.execute(
                    "SELECT COUNT(*) FROM catalog WHERE resource_path = %s AND resource_version = %s",
                    (resource_path, resource_version)
                )
                count = int(cursor.fetchone()[0])  # Convert to int
                max_attempts = 10
                attempt = 1
                while count > 0 and attempt <= max_attempts:
                    logger.warning(f"Version '{resource_version}' already exists for resource_path '{resource_path}', incrementing again (attempt {attempt}/{max_attempts})")
                    resource_version = self.increment_version(resource_version)
                    logger.debug(f"Incremented version: '{resource_version}'")

                    cursor.execute(
                        "SELECT COUNT(*) FROM catalog WHERE resource_path = %s AND resource_version = %s",
                        (resource_path, resource_version)
                    )
                    count = int(cursor.fetchone()[0])  # Convert to int
                    attempt += 1

                if attempt > max_attempts:
                    logger.error(f"Failed to find version after {max_attempts} attempts")
                    raise HTTPException(
                        status_code=500,
                        detail=f"Failed to find version after {max_attempts} attempts"
                    )

            logger.info(f"Registering resource '{resource_path}' with version '{resource_version}' (previous: '{latest_version}')")

            with conn.cursor() as cursor:
                cursor.execute(
                    "INSERT INTO resource (name) VALUES (%s) ON CONFLICT DO NOTHING",
                    (resource_type,)
                )
                cursor.execute(
                    """
                    INSERT INTO catalog 
                    (resource_path, resource_type, resource_version, content, payload, meta)
                    VALUES (%s, %s, %s, %s, %s, %s)
                    """,
                    (
                        resource_path,
                        resource_type,
                        resource_version,
                        content,
                        json.dumps(resource_data),
                        json.dumps({"registered_at": "now()"})
                    )
                )
                conn.commit()

            return {
                "status": "success",
                "message": f"Resource '{resource_path}' version '{resource_version}' registered.",
                "resource_path": resource_path,
                "resource_version": resource_version,
                "resource_type": resource_type
            }

        except Exception as e:
            logger.exception(f"Error registering resource: {e}.")
            if conn:
                try:
                    conn.rollback()
                    logger.info("Transaction rolled back due to error.")
                except Exception as rollback_error:
                    logger.exception(f"Error rolling back transaction: {rollback_error}.")
            raise HTTPException(
                status_code=500,
                detail=f"Error registering resource: {e}."
            )
        finally:
            if conn:
                try:
                    conn.close()
                    logger.debug("Connection closed.")
                except Exception as close_error:
                    logger.exception(f"Error closing connection: {close_error}.")

    def list_entries(self, resource_type: Optional[str] = None) -> List[Dict[str, Any]]:
        conn = None
        try:
            conn = psycopg.connect(self.pgdb_conn_string)

            with conn.cursor() as cursor:
                if resource_type:
                    cursor.execute(
                        """
                        SELECT resource_path, resource_type, resource_version, meta, timestamp
                        FROM catalog
                        WHERE resource_type = %s
                        ORDER BY timestamp DESC
                        """,
                        (resource_type,)
                    )
                else:
                    cursor.execute(
                        """
                        SELECT resource_path, resource_type, resource_version, meta, timestamp
                        FROM catalog
                        ORDER BY timestamp DESC
                        """
                    )

                results = cursor.fetchall()

            entries = []
            for row in results:
                entries.append({
                    "resource_path": row[0],
                    "resource_type": row[1],
                    "resource_version": row[2],
                    "meta": row[3],
                    "timestamp": row[4]
                })

            return entries

        except Exception as e:
            logger.exception(f"Error listing catalog entries: {e}")
            return []
        finally:
            if conn:
                try:
                    conn.close()
                except Exception as close_error:
                    logger.exception(f"Error closing connection: {close_error}")



def get_catalog_service() -> CatalogService:
    return CatalogService()

class EventService:
    def __init__(self, pgdb_conn_string: str | None = None):
        self.pgdb_conn_string = pgdb_conn_string if pgdb_conn_string else get_pgdb_connection()

    def emit(self, event_data: Dict[str, Any]) -> Dict[str, Any]:
        conn = None
        try:
            event_id = event_data.get("event_id", f"evt_{os.urandom(16).hex()}")
            event_data["event_id"] = event_id
            event_type = event_data.get("event_type", "UNKNOWN")
            status = event_data.get("status", "CREATED")
            parent_event_id = event_data.get("parent_id") or event_data.get("parent_event_id")
            execution_id = event_data.get("execution_id", event_id)  # Use event_id as execution_id if not provided
            node_id = event_data.get("node_id", event_id)
            node_name = event_data.get("node_name", event_type)
            node_type = event_data.get("node_type", "event")
            duration = event_data.get("duration", 0.0)
            metadata = event_data.get("meta", {})
            error = event_data.get("error")
            input_context = json.dumps(event_data.get("context", {}))
            output_result = json.dumps(event_data.get("result", {}))
            metadata_str = json.dumps(metadata)

            conn = psycopg.connect(self.pgdb_conn_string)
            with conn.cursor() as cursor:
                cursor.execute("""
                    SELECT COUNT(*) FROM event_log 
                    WHERE execution_id = %s AND event_id = %s
                """, (execution_id, event_id))

                exists = cursor.fetchone()[0] > 0

                if exists:
                    cursor.execute("""
                        UPDATE event_log SET
                            event_type = %s,
                            status = %s,
                            duration = %s,
                            input_context = %s,
                            output_result = %s,
                            metadata = %s,
                            error = %s,
                            timestamp = CURRENT_TIMESTAMP
                        WHERE execution_id = %s AND event_id = %s
                    """, (
                        event_type,
                        status,
                        duration,
                        input_context,
                        output_result,
                        metadata_str,
                        error,
                        execution_id,
                        event_id
                    ))
                else:
                    cursor.execute("""
                        INSERT INTO event_log (
                            execution_id, event_id, parent_event_id, timestamp, event_type,
                            node_id, node_name, node_type, status, duration,
                            input_context, output_result, metadata, error
                        ) VALUES (
                            %s, %s, %s, CURRENT_TIMESTAMP, %s,
                            %s, %s, %s, %s, %s,
                            %s, %s, %s, %s
                        )
                    """, (
                        execution_id,
                        event_id,
                        parent_event_id,
                        event_type,
                        node_id,
                        node_name,
                        node_type,
                        status,
                        duration,
                        input_context,
                        output_result,
                        metadata_str,
                        error
                    ))

                conn.commit()

            logger.info(f"Event emitted: {event_id} - {event_type} - {status}")
            return event_data

        except Exception as e:
            logger.exception(f"Error emitting event: {e}")
            raise HTTPException(
                status_code=500,
                detail=f"Error emitting event: {e}"
            )
        finally:
            if conn:
                try:
                    conn.close()
                except Exception as close_error:
                    logger.exception(f"Error closing connection: {close_error}")

    def get_events_by_execution_id(self, execution_id: str) -> Optional[Dict[str, Any]]:
        """
        Get all events for a specific execution.

        Args:
            execution_id: The ID of the execution

        Returns:
            A dictionary containing events or None if not found
        """
        conn = None
        try:
            conn = psycopg.connect(self.pgdb_conn_string)
            with conn.cursor() as cursor:
                cursor.execute("""
                    SELECT 
                        event_id, 
                        event_type, 
                        node_id, 
                        node_name, 
                        node_type, 
                        status, 
                        duration, 
                        timestamp, 
                        input_context, 
                        output_result, 
                        metadata, 
                        error
                    FROM event_log 
                    WHERE execution_id = %s
                    ORDER BY timestamp
                """, (execution_id,))

                rows = cursor.fetchall()
                if rows:
                    events = []
                    for row in rows:
                        event_data = {
                            "event_id": row[0],
                            "event_type": row[1],
                            "node_id": row[2],
                            "node_name": row[3],
                            "node_type": row[4],
                            "status": row[5],
                            "duration": row[6],
                            "timestamp": row[7].isoformat() if row[7] else None,
                            "input_context": json.loads(row[8]) if row[8] else None,
                            "output_result": json.loads(row[9]) if row[9] else None,
                            "metadata": json.loads(row[10]) if row[10] else None,
                            "error": row[11],
                            "execution_id": execution_id,
                            "resource_path": None,
                            "resource_version": None
                        }

                        # Try to extract resource_path and resource_version from metadata or input_context
                        if event_data["metadata"] and "playbook_path" in event_data["metadata"]:
                            event_data["resource_path"] = event_data["metadata"]["playbook_path"]

                        if event_data["input_context"] and "path" in event_data["input_context"]:
                            event_data["resource_path"] = event_data["input_context"]["path"]

                        if event_data["input_context"] and "version" in event_data["input_context"]:
                            event_data["resource_version"] = event_data["input_context"]["version"]

                        events.append(event_data)

                    return {"events": events}

                return None
        except Exception as e:
            logger.exception(f"Error getting events by execution_id: {e}")
            return None
        finally:
            if conn:
                try:
                    conn.close()
                except Exception as close_error:
                    logger.exception(f"Error closing connection: {close_error}")

    def get_event_by_id(self, event_id: str) -> Optional[Dict[str, Any]]:
        """
        Get a single event by its ID.

        Args:
            event_id: The ID of the event

        Returns:
            A dictionary containing the event or None if not found
        """
        conn = None
        try:
            conn = psycopg.connect(self.pgdb_conn_string)
            with conn.cursor() as cursor:
                cursor.execute("""
                    SELECT 
                        event_id, 
                        event_type, 
                        node_id, 
                        node_name, 
                        node_type, 
                        status, 
                        duration, 
                        timestamp, 
                        input_context, 
                        output_result, 
                        metadata, 
                        error,
                        execution_id
                    FROM event_log 
                    WHERE event_id = %s
                """, (event_id,))

                row = cursor.fetchone()
                if row:
                    event_data = {
                        "event_id": row[0],
                        "event_type": row[1],
                        "node_id": row[2],
                        "node_name": row[3],
                        "node_type": row[4],
                        "status": row[5],
                        "duration": row[6],
                        "timestamp": row[7].isoformat() if row[7] else None,
                        "input_context": json.loads(row[8]) if row[8] else None,
                        "output_result": json.loads(row[9]) if row[9] else None,
                        "metadata": json.loads(row[10]) if row[10] else None,
                        "error": row[11],
                        "execution_id": row[12],
                        "resource_path": None,
                        "resource_version": None
                    }
                    if event_data["metadata"] and "playbook_path" in event_data["metadata"]:
                        event_data["resource_path"] = event_data["metadata"]["playbook_path"]

                    if event_data["input_context"] and "path" in event_data["input_context"]:
                        event_data["resource_path"] = event_data["input_context"]["path"]

                    if event_data["input_context"] and "version" in event_data["input_context"]:
                        event_data["resource_version"] = event_data["input_context"]["version"]
                    return {"events": [event_data]}

                return None
        except Exception as e:
            logger.exception(f"Error getting event by ID: {e}")
            return None
        finally:
            if conn:
                try:
                    conn.close()
                except Exception as close_error:
                    logger.exception(f"Error closing connection: {close_error}")

    def get_event(self, id_param: str) -> Optional[Dict[str, Any]]:
        """
        Get events by execution_id or event_id (legacy method for backward compatibility).

        Args:
            id_param: Either an execution_id or an event_id

        Returns:
            A dictionary containing events or None if not found
        """
        conn = None
        try:
            conn = psycopg.connect(self.pgdb_conn_string)
            with conn.cursor() as cursor:
                cursor.execute("""
                    SELECT COUNT(*) FROM event_log WHERE execution_id = %s
                """, (id_param,))
                count = cursor.fetchone()[0]

                if count > 0:
                    events = self.get_events_by_execution_id(id_param)
                    if events:
                        return events

            event = self.get_event_by_id(id_param)
            if event:
                return event

            with conn.cursor() as cursor:
                cursor.execute("""
                    SELECT DISTINCT execution_id FROM event_log 
                    WHERE event_id = %s
                """, (id_param,))
                execution_ids = [row[0] for row in cursor.fetchall()]

                if execution_ids:
                    events = self.get_events_by_execution_id(execution_ids[0])
                    if events:
                        return events

            return None
        except Exception as e:
            logger.exception(f"Error in get_event: {e}")
            return None
        finally:
            if conn:
                try:
                    conn.close()
                except Exception as close_error:
                    logger.exception(f"Error closing connection: {close_error}")

def get_event_service() -> EventService:
    return EventService(get_pgdb_connection())


class AgentService:

    def __init__(self, pgdb_conn_string: str | None = None):
        self.pgdb_conn_string = pgdb_conn_string if pgdb_conn_string else get_pgdb_connection()
        self.agent = None

    def store_transition(self, params: tuple):
        """
        Store the transition in the database.

        Args:
            params: A tuple containing the transition parameters
        """
        if self.agent:
            self.agent.store_transition(params)

    def get_step_results(self) -> Dict[str, Any]:
        """
        Get the results of all steps.

        Returns:
            A dictionary mapping the step names to results
        """
        if self.agent:
            return self.agent.get_step_results()
        return {}

    def execute_agent(
        self, 
        playbook_content: str, 
        playbook_path: str, 
        playbook_version: str, 
        input_payload: Optional[Dict[str, Any]] = None,
        sync_to_postgres: bool = True,
        merge: bool = False
    ) -> Dict[str, Any]:
        try:
            with tempfile.NamedTemporaryFile(suffix=".yaml", delete=False) as temp_file:
                temp_file.write(playbook_content.encode('utf-8'))
                temp_file_path = temp_file.name
            try:
                pgdb_conn = self.pgdb_conn_string if sync_to_postgres else None
                self.agent = NoETLAgent(temp_file_path, mock_mode=False, pgdb=pgdb_conn)
                agent = self.agent
                workload = agent.playbook.get('workload', {})
                if input_payload:
                    if merge:
                        logger.info("Merge mode: deep merging input payload with workload.")
                        merged_workload = deep_merge(workload, input_payload)
                        for key, value in merged_workload.items():
                            agent.update_context(key, value)
                        agent.update_context('workload', merged_workload)
                        agent.store_workload(merged_workload)
                    else:
                        logger.info("Override mode: replacing workload keys with input payload.")
                        merged_workload = workload.copy()
                        for key, value in input_payload.items():
                            merged_workload[key] = value
                        for key, value in merged_workload.items():
                            agent.update_context(key, value)
                        agent.update_context('workload', merged_workload)
                        agent.store_workload(merged_workload)
                else:
                    logger.info("No input payload provided. Default workload from playbook is used.")
                    for key, value in workload.items():
                        agent.update_context(key, value)
                    agent.update_context('workload', workload)
                    agent.store_workload(workload)
                server_url = os.environ.get('NOETL_SERVER_URL', 'http://localhost:8082')
                daemon = Broker(agent, server_url=server_url)
                results = daemon.run()

                export_path = None

                return {
                    "status": "success",
                    "message": f"Agent executed for playbook '{playbook_path}' version '{playbook_version}'.",
                    "result": results,
                    "execution_id": agent.execution_id,
                    "export_path": export_path
                }
            finally:
                if os.path.exists(temp_file_path):
                    os.unlink(temp_file_path)

        except Exception as e:
            logger.exception(f"Error executing agent: {e}.")
            return {
                "status": "error",
                "message": f"Error executing agent for playbook '{playbook_path}' version '{playbook_version}': {e}.",
                "error": str(e)
            }

def get_agent_service() -> AgentService:
    return AgentService(get_pgdb_connection())

@router.post("/catalog/register", response_class=JSONResponse)
async def register_resource(
    request: Request,
    content_base64: str = None,
    content: str = None,
    resource_type: str = "playbook"
):
    try:
        if not content_base64 and not content:
            try:
                body = await request.json()
                content_base64 = body.get("content_base64")
                content = body.get("content")
                resource_type = body.get("resource_type", resource_type)
            except:
                pass

        if content_base64:
            import base64
            content = base64.b64decode(content_base64).decode('utf-8')
        elif not content:
            raise HTTPException(
                status_code=400,
                detail="The content or content_base64 must be provided."
            )

        catalog_service = get_catalog_service()
        result = catalog_service.register_resource(content, resource_type)
        return result

    except Exception as e:
        logger.exception(f"Error registering resource: {e}.")
        raise HTTPException(
            status_code=500,
            detail=f"Error registering resource: {e}."
        )

@router.get("/catalog/list", response_class=JSONResponse)
async def list_resources(
    request: Request,
    resource_type: str = None
):
    try:
        catalog_service = get_catalog_service()
        entries = catalog_service.list_entries(resource_type)
        return {"entries": entries}

    except Exception as e:
        logger.exception(f"Error listing resources: {e}")
        raise HTTPException(
            status_code=500,
            detail=f"Error listing resources: {e}"
        )

@router.get("/catalog/{path:path}/{version}", response_class=JSONResponse)
async def get_resource(
    request: Request,
    path: str,
    version: str
):
    try:
        catalog_service = get_catalog_service()
        entry = catalog_service.fetch_entry(path, version)
        if not entry:
            raise HTTPException(
                status_code=404,
                detail=f"Resource '{path}' with version '{version}' not found."
            )
        return entry

    except HTTPException:
        raise
    except Exception as e:
        logger.exception(f"Error fetching resource: {e}.")
        raise HTTPException(
            status_code=500,
            detail=f"Error fetching resource: {e}."
        )


@router.get("/events/by-execution/{execution_id}", response_class=JSONResponse)
async def get_events_by_execution(
    request: Request,
    execution_id: str
):
    """
    Get all events for a specific execution.
    """
    try:
        event_service = get_event_service()
        events = event_service.get_events_by_execution_id(execution_id)
        if not events:
            raise HTTPException(
                status_code=404,
                detail=f"No events found for execution '{execution_id}'."
            )
        return events

    except HTTPException:
        raise
    except Exception as e:
        logger.exception(f"Error fetching events by execution: {e}")
        raise HTTPException(
            status_code=500,
            detail=f"Error fetching events by execution: {e}"
        )

@router.get("/events/by-id/{event_id}", response_class=JSONResponse)
async def get_event_by_id(
    request: Request,
    event_id: str
):
    """
    Get a single event by its ID.
    """
    try:
        event_service = get_event_service()
        event = event_service.get_event_by_id(event_id)
        if not event:
            raise HTTPException(
                status_code=404,
                detail=f"Event with ID '{event_id}' not found."
            )
        return event

    except HTTPException:
        raise
    except Exception as e:
        logger.exception(f"Error fetching event by ID: {e}")
        raise HTTPException(
            status_code=500,
            detail=f"Error fetching event by ID: {e}"
        )

@router.get("/events/{event_id}", response_class=JSONResponse)
async def get_event(
    request: Request,
    event_id: str
):
    """
    Legacy endpoint for getting events by execution_id or event_id.
    Use /events/by-execution/{execution_id} or /events/by-id/{event_id} instead.
    """
    try:
        event_service = get_event_service()
        event = event_service.get_event(event_id)
        if not event:
            raise HTTPException(
                status_code=404,
                detail=f"Event '{event_id}' not found."
            )
        return event

    except HTTPException:
        raise
    except Exception as e:
        logger.exception(f"Error fetching event: {e}")
        raise HTTPException(
            status_code=500,
            detail=f"Error fetching event: {e}"
        )

@router.get("/events/query", response_class=JSONResponse)
async def get_event_by_query(
    request: Request,
    event_id: str = None
):
    if not event_id:
        raise HTTPException(
            status_code=400,
            detail="event_id query parameter is required."
        )

    try:
        event_service = get_event_service()
        event = event_service.get_event(event_id)
        if not event:
            raise HTTPException(
                status_code=404,
                detail=f"Event '{event_id}' not found."
            )
        return event

    except HTTPException:
        raise
    except Exception as e:
        logger.exception(f"Error fetching event: {e}.")
        raise HTTPException(
            status_code=500,
            detail=f"Error fetching event: {e}."
        )

@router.get("/execution/data/{execution_id}", response_class=JSONResponse)
async def get_execution_data(
    request: Request,
    execution_id: str
):
    try:
        event_service = get_event_service()
        event = event_service.get_event(execution_id)
        if not event:
            raise HTTPException(
                status_code=404,
                detail=f"Execution '{execution_id}' not found."
            )
        return event

    except HTTPException:
        raise
    except Exception as e:
        logger.exception(f"Error fetching execution data: {e}.")
        raise HTTPException(
            status_code=500,
            detail=f"Error fetching execution data: {e}."
        )

@router.post("/events", response_class=JSONResponse)
async def create_event(
    request: Request
):
    try:
        body = await request.json()
        event_service = get_event_service()
        result = event_service.emit(body)
        return result
    except Exception as e:
        logger.exception(f"Error creating event: {e}.")
        raise HTTPException(
            status_code=500,
            detail=f"Error creating event: {e}."
        )


@router.post("/agent/execute", response_class=JSONResponse)
async def execute_agent(
    request: Request,
    path: str = None,
    version: str = None,
    input_payload: Dict[str, Any] = None,
    sync_to_postgres: bool = True,
    merge: bool = False
):
    try:
        if not path:
            try:
                body = await request.json()
                path = body.get("path", path)
                version = body.get("version", version)
                input_payload = body.get("input_payload", input_payload)
                sync_to_postgres = body.get("sync_to_postgres", sync_to_postgres)
                merge = body.get("merge", merge)
            except:
                pass

        if not path:
            raise HTTPException(
                status_code=400,
                detail="Path is a required parameter."
            )

        catalog_service = get_catalog_service()
        if not version:
            version = catalog_service.get_latest_version(path)
            logger.info(f"Version not specified, using latest version: {version}")

        entry = catalog_service.fetch_entry(path, version)
        if not entry:
            raise HTTPException(
                status_code=404,
                detail=f"Playbook '{path}' with version '{version}' not found."
            )

        agent_service = get_agent_service()
        result = agent_service.execute_agent(
            playbook_content=entry.get("content"),
            playbook_path=path,
            playbook_version=version,
            input_payload=input_payload,
            sync_to_postgres=sync_to_postgres,
            merge=merge
        )

        return result

    except Exception as e:
        logger.exception(f"Error executing agent: {e}")
        raise HTTPException(
            status_code=500,
            detail=f"Error executing agent for playbook '{path}' version '{version}': {e}."
        )

@router.post("/agent/execute-async", response_class=JSONResponse)
async def execute_agent_async(
    request: Request,
    background_tasks: BackgroundTasks,
    path: str = None,
    version: str = None,
    input_payload: Dict[str, Any] = None,
    sync_to_postgres: bool = True,
    merge: bool = False
):

    try:
        if not path:
            try:
                body = await request.json()
                path = body.get("path", path)
                version = body.get("version", version)
                input_payload = body.get("input_payload", input_payload)
                sync_to_postgres = body.get("sync_to_postgres", sync_to_postgres)
                merge = body.get("merge", merge)
            except:
                pass

        if not path:
            raise HTTPException(
                status_code=400,
                detail="Path is a required parameter."
            )

        catalog_service = get_catalog_service()

        if not version:
            version = catalog_service.get_latest_version(path)
            logger.info(f"Version not specified, using latest version: {version}")

        entry = catalog_service.fetch_entry(path, version)
        if not entry:
            raise HTTPException(
                status_code=404,
                detail=f"Playbook '{path}' with version '{version}' not found."
            )

        event_service = get_event_service()
        initial_event_data = {
            "event_type": "AgentExecutionRequested",
            "status": "REQUESTED",
            "meta": {
                "resource_path": path,
                "resource_version": version,
            },
            "result": input_payload,
            "node_type": "playbook",
            "node_name": path
        }

        initial_event = event_service.emit(initial_event_data)

        def execute_agent_task():
            try:
                with tempfile.NamedTemporaryFile(suffix=".yaml", delete=False) as temp_file:
                    temp_file.write(entry.get("content").encode('utf-8'))
                    temp_file_path = temp_file.name

                try:
                    pgdb_conn = get_pgdb_connection() if sync_to_postgres else None
                    agent = NoETLAgent(temp_file_path, mock_mode=False, pgdb=pgdb_conn)
                    workload = agent.playbook.get('workload', {})
                    if input_payload:
                        if merge:
                            logger.info("Merge mode: deep merging input payload with workload.")
                            merged_workload = deep_merge(workload, input_payload)

                            for key, value in merged_workload.items():
                                agent.update_context(key, value)

                            agent.update_context('workload', merged_workload)
                            agent.store_workload(merged_workload)
                        else:
                            logger.info("Override mode: replacing workload keys with input payload.")
                            merged_workload = workload.copy()
                            for key, value in input_payload.items():
                                merged_workload[key] = value
                            for key, value in merged_workload.items():
                                agent.update_context(key, value)
                            agent.update_context('workload', merged_workload)
                            agent.store_workload(merged_workload)
                    else:
                        logger.info("No input payload provided. Default workload from playbook is used.")

                        for key, value in workload.items():
                            agent.update_context(key, value)

                        agent.update_context('workload', workload)
                        agent.store_workload(workload)

                    results = agent.run()
                    event_id = initial_event.get("event_id")
                    update_event = {
                        "event_id": event_id,
                        "execution_id": agent.execution_id,
                        "event_type": "agent_execution_completed",
                        "status": "COMPLETED",
                        "result": results,
                        "meta": {
                            "resource_path": path,
                            "resource_version": version,
                            "execution_id": agent.execution_id
                        },
                        "node_type": "playbook",
                        "node_name": path
                    }

                    event_service.emit(update_event)
                    logger.info(f"Event updated: {event_id} - agent_execution_completed - COMPLETED")

                finally:
                    if os.path.exists(temp_file_path):
                        os.unlink(temp_file_path)

            except Exception as e:
                logger.exception(f"Error in background agent execution: {e}.")
                event_id = initial_event.get("event_id")

                error_event = {
                    "event_id": event_id,
                    "execution_id": event_id,
                    "event_type": "agent_execution_error",
                    "status": "ERROR",
                    "error": str(e),
                    "result": {"error": str(e)},
                    "meta": {
                        "resource_path": path,
                        "resource_version": version,
                        "error": str(e)
                    },
                    "node_type": "playbook",
                    "node_name": path
                }

                event_service.emit(error_event)
                logger.info(f"Event updated: {event_id} - agent_execution_error - ERROR")

        background_tasks.add_task(execute_agent_task)

        return {
            "status": "accepted",
            "message": f"Agent execution started for playbook '{path}' version '{version}'.",
            "event_id": initial_event.get("event_id")
        }

    except Exception as e:
        logger.exception(f"Error starting agent execution: {e}.")
        raise HTTPException(
            status_code=500,
            detail=f"Error starting agent execution for playbook '{path}' version '{version}': {e}"
        )
